"""Test whether Skyfield handles file download and file age correctly."""

import os
import shutil
import tempfile
from contextlib import contextmanager
from threading import Thread
try:
    from Queue import Queue
except ImportError:
    from queue import Queue
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from assay import assert_raises
from skyfield import api

data_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '../data'))
old_content = (b' 2015 10  1  67.9546\n'
               b' 2015 11  1  68.0055\n'
               b' 2015 12  1  68.0514\n'
               b' 2016  1  1  68.1024\n')
new_content = (old_content +
               b' 2016  2  1  68.1577\n')

def load():
    path = tempfile.mkdtemp()
    try:
        yield api.Loader(path)
    finally:
        shutil.rmtree(path)

def save_file(load, path, content=old_content):
    with open(load.path_to(path), 'wb') as f:
        f.write(content)

def file_contents(load, path):
    with open(load.path_to(path), 'rb') as f:
        return f.read()

@contextmanager
def fake_download(load):
    download = lambda *args, **kw: save_file(load, 'deltat.data', new_content)
    with patch('skyfield.iokit.download', download):
        yield

# Simple tests.

def test_build_url(load):
    url = 'ftp://ssd.jpl.nasa.gov/pub/eph/planets/bsp/de421.bsp'
    assert load.build_url('de421.bsp') == url
    with assert_raises(ValueError, 'know the URL'):
        load.build_url('unknown.kind.of.file')

def test_open_in_main_directory(load):
    with open(os.path.join(load.directory, 'file.tle'), 'wb') as f:
        f.write(b'example text\n')
    data = load.open('file.tle').read()
    assert data == b'example text\n'

def test_open_in_subdirectory(load):
    os.mkdir(os.path.join(load.directory, 'folder'))
    with open(os.path.join(load.directory, 'folder', 'file.tle'), 'wb') as f:
        f.write(b'example text\n')
    data = load.open('folder/file.tle').read()
    assert data == b'example text\n'

def test_missing_file_gets_downloaded(load):
    with fake_download(load):
        data = load('deltat.data')
    assert file_contents(load, 'deltat.data').endswith(b' 68.1577\n')
    assert data[1][-1] == 68.1577

def test_builtin_timescale_uses_recent_IERS_data(load):
    ts = load.timescale()
    # DUT1 cut and pasted from "20 1 1" line of "finals2000A.all":
    assert abs(ts.utc(2020, 1, 1).dut1 - (-0.1771547)) < 1e-8

def test_non_builtin_timescale_prefers_USNO_files(load):
    with open(os.path.join(data_dir, 'deltat.preds'), 'rb') as f:
        preds = f.read()
    with open(os.path.join(data_dir, 'Leap_Second.dat'), 'rb') as f:
        leaps = f.read()

    save_file(load, 'deltat.data',
              b' 1973  2  1  111.1\n'
              b' 1973  3  1  222.2\n'
              b' 1973  4  1  333.3\n')
    save_file(load, 'deltat.preds', preds)
    save_file(load, 'Leap_Second.dat', leaps)
    save_file(load, 'finals2000A.all', b'invalid data')

    ts = load.timescale(builtin=False)
    assert abs(ts.utc(1973, 3, 1).delta_t - 222.2) < 1e-2

def test_non_builtin_timescale_tries_to_load_finals2000A_all(load):
    save_file(load, 'finals2000A.all', b'invalid data')
    with assert_raises(IndexError):
        load.timescale(builtin=False)

# Impressive tests: synchronize threads to reproduce concurrency bugs.

class FakeConnection():
    def __init__(self):
        self.control = Queue()
        self.blocks = Queue()
        self.headers = {}

    def __call__(self, *args, **kw):  # Intercept urlopen()
        return self

    def read(self, blocksize):  # Intercept connection.read()
        self.control.put('Read requested')
        return self.blocks.get()

def test_concurrent_downloads(load):
    # While downloads in a single process are protecte

    eof = b''
    data = b' 1973  2  1  43.4724\n'

    c1 = FakeConnection()
    c2 = FakeConnection()
    t1 = Thread(target=load, args=('deltat.data',))
    t2 = Thread(target=load, args=('deltat.data',))
    t1.daemon = True
    t2.daemon = True

    with patch('skyfield.iokit.urlopen', c1):
        t1.start()
        c1.control.get()  # has reached its first read()
    with patch('skyfield.iokit.urlopen', c2):
        t2.start()
        c2.control.get()  # has reached its first read()

    assert sorted(os.listdir(load.directory)) == [
        'deltat.data.download', 'deltat.data.download2',
    ]
    c1.blocks.put(data)
    c1.blocks.put(eof)
    t1.join()
    assert sorted(os.listdir(load.directory)) == [
        'deltat.data', 'deltat.data.download2',
    ]
    c2.blocks.put(data)
    c2.blocks.put(eof)
    t2.join()
    assert sorted(os.listdir(load.directory)) == ['deltat.data']



from __future__ import print_function, division
from skyfield import api

def test_sat_almanac_LEO():
    # Testcase from
    # https://github.com/skyfielders/astronomy-notebooks/blob/master/Solvers/Earth-Satellite-Passes.ipynb
    # with elevated horizon

    tle = ["TIANGONG 1",
           "1 37820U 11053A   14314.79851609  .00064249  00000-0  44961-3 0  5637",
           "2 37820  42.7687 147.7173 0010686 283.6368 148.1694 15.73279710179072"]
    sat = api.EarthSatellite(*tle[1:3], name=tle[0])
    topos = api.Topos('42.3581 N', '71.0636 W')
    timescale = api.load.timescale()
    t0 = timescale.tai(2014, 11, 10)
    t1 = timescale.tai(2014, 11, 11)
    horizon = 20
    nexpected = 12

    #times, yis = almanac.find_satellite_events(t0, t1, sat, topos, horizon=20)
    times, yis = sat.find_events(topos, t0, t1, 20.0)
    assert(verify_sat_almanac(times, yis, sat, topos, horizon, nexpected))


def test_sat_almanac_Tricky():
    # Various tricky satellites
    # Integral: 3 days high eccentricity.
    # ANIK-F1R  Geo always visible from Boston
    # PALAPA D  Geo never visible from Boston
    # Ariane 5B GTO
    # Swift Low-inclination LEO never visible from Boston
    # Grace-FO 2  Low polar orbit
    tles = """\
        INTEGRAL
        1 27540U 02048A   20007.25125384  .00001047  00000-0  00000+0 0  9992
        2 27540  51.8988 127.5680 8897013 285.8757   2.8911  0.37604578 17780
        ANIK F-1R
        1 28868U 05036A   20011.46493281 -.00000066  00000-0  00000+0 0  9999
        2 28868   0.0175  50.4632 0002403 284.1276 195.8977  1.00270824 52609
        PALAPA D
        1 35812U 09046A   20008.38785173 -.00000341 +00000-0 +00000-0 0  9999
        2 35812 000.0518 095.9882 0002721 218.8296 045.1595 01.00269700038098
        Ariane 5B
        1 44802U 19080C   20010.68544515  .00001373  00000-0  27860-3 0  9997
        2 44802   5.1041 192.7327 7266711 217.6622  57.0965  2.30416801  1028
        Swift
        1 28485U 04047A   20010.76403232 +.00000826 +00000-0 +25992-4 0  9999
        2 28485 020.5579 055.7027 0010957 208.9479 151.0347 15.04516653829549
        GRACE-FO 2
        1 43477U 18047B   20011.66650462 +.00000719  00000-0 +29559-4 0    08
        2 43477  88.9974 159.0391 0019438 141.4770 316.8932 15.23958285 91199
    """.splitlines()
    expected = {'INTEGRAL': 36,
                'ANIK F-1R': 14,
                'PALAPA D': 0,
                'Ariane 5B': 37,
                'Swift': 0,
                'GRACE-FO 2': 90}
    for iline0 in range(0, len(tles)-3, 3):
        sat = api.EarthSatellite(tles[1+iline0].strip(),
                                 tles[2+iline0].strip(),
                                 name=tles[iline0].strip())
        topos = api.Topos('42.3581 N', '71.0636 W')  # Boston
        timescale = api.load.timescale()
        t0 = timescale.tai(2020, 1, 1)
        t1 = timescale.tai(2020, 1, 15)
        horizon = 20
        nexpected = expected[sat.name]

        # times, yis = almanac.find_satellite_events(t0, t1, sat, topos, horizon=20)
        times, yis = sat.find_events(topos, t0, t1, 20.0)
        assert(verify_sat_almanac(times, yis, sat, topos, horizon, nexpected))



# Helper function to verify satellite events
def verify_sat_almanac(times, yis, sat, topos, horizon, nexpected):
    if nexpected is None:
        print("Number of satellite events expected for", sat.name," not specified.  Got ", len(times))
    else:
        if len(times) != nexpected or len(yis) != nexpected:
            raise RuntimeError("Expected {} events for satellite {}"
                               " but got {} times and {} y values"
                               .format(nexpected, sat.model.satnum,
                                       len(times), len(yis)))
    if len(times) == 0:
        return True     # Nothing to check
    # Verify 1) rises/sets cross the horizon in the right direction
    # 2) culminations are local maxima above horizon
    # 3) No double rises or double sets
    time_tolerance = 5/(24 * 60 * 60)   # Times must be accurate within 5 seconds
    ts = times[0].ts     # Use the timescale passed by
    altitudes_neartimes = [
        (sat - topos).at(ts.tai(jd=times.tai + dt)).altaz()[0].degrees
        for dt in (-time_tolerance, 0, time_tolerance)
    ]
    lastevent = None
    event_dict = {0: 'rise', 1: 'culminate', 2: 'set'}
    for time, yi, (alt_before, alt_at, alt_after) in zip(
            times, yis, zip(*altitudes_neartimes)):
        eventname = event_dict[yi]
        if eventname == 'rise':
            assert(alt_before < alt_at < alt_after)  # It is going up
            assert(alt_before < horizon < alt_after) # It is crossing horizon altitude
            assert(lastevent != 'rise')     # Can't have two consecutive rises
        elif eventname == 'culminate':
            assert(alt_before < alt_at)
            assert(alt_at > alt_after)
            assert(alt_at > horizon)
            # You can have two consecutive culminations (esp. for GEOs).
        elif eventname == 'set':
            assert(alt_before > alt_at > alt_after)  # It is going down
            assert(alt_before > horizon > alt_after) # It is crossing horizon altitude
            assert(lastevent != 'set')     # Can't have two consecutive sets
        else:
            raise RuntimeError("Unexpected satellite event type")
        lastevent = eventname
    return True



"""Tests of whether units behave."""

from assay import assert_raises
from numpy import array
from skyfield.units import Angle, Distance, Velocity, UnpackingError

try:
    from astropy import units as u
except ImportError:
    u = None

def needs_astropy(test):
    """Skip `test` if AstroPy is not available."""
    return None if (u is None) else test

def test_degree_rounding():
    tenth = 0.1 / 60.0 / 60.0  # of an arcsecond

    assert str(Angle(degrees=tenth * -600.75)) == '-00deg 01\' 00.1"'
    assert str(Angle(degrees=tenth * -600.25)) == '-00deg 01\' 00.0"'
    assert str(Angle(degrees=tenth * -599.75)) == '-00deg 01\' 00.0"'
    assert str(Angle(degrees=tenth * -599.25)) == '-00deg 00\' 59.9"'

    assert str(Angle(degrees=tenth * -1.75)) == '-00deg 00\' 00.2"'
    assert str(Angle(degrees=tenth * -1.25)) == '-00deg 00\' 00.1"'
    assert str(Angle(degrees=tenth * -0.75)) == '-00deg 00\' 00.1"'
    assert str(Angle(degrees=tenth * -0.25)) == '-00deg 00\' 00.0"'

    assert str(Angle(degrees=0.0)) == '00deg 00\' 00.0"'

    assert str(Angle(degrees=tenth * 0.25)) == '00deg 00\' 00.0"'
    assert str(Angle(degrees=tenth * 0.75)) == '00deg 00\' 00.1"'
    assert str(Angle(degrees=tenth * 1.25)) == '00deg 00\' 00.1"'
    assert str(Angle(degrees=tenth * 1.75)) == '00deg 00\' 00.2"'

    assert str(Angle(degrees=tenth * 599.25)) == '00deg 00\' 59.9"'
    assert str(Angle(degrees=tenth * 599.75)) == '00deg 01\' 00.0"'
    assert str(Angle(degrees=tenth * 600.25)) == '00deg 01\' 00.0"'
    assert str(Angle(degrees=tenth * 600.75)) == '00deg 01\' 00.1"'

def test_angle_scalar_strs():
    assert str(Angle(degrees=array(91))) == '''91deg 00' 00.0"'''
    assert str(Angle(hours=array(12))) == '''12h 00m 00.00s'''

def test_angle_array_strs():
    assert str(Angle(degrees=array([90, 91, 92]))) == (
        '''3 values from 90deg 00' 00.0" to 92deg 00' 00.0"'''
        )
    assert str(Angle(hours=array([11, 12, 13]))) == (
        '''3 values from 11h 00m 00.00s to 13h 00m 00.00s'''
        )

def test_angle_sexagesimal_args():
    assert str(Angle(degrees=(90,))) == '''90deg 00' 00.0"'''
    assert str(Angle(hours=(12,))) == '''12h 00m 00.00s'''

    assert str(Angle(degrees=(90, 15))) == '''90deg 15' 00.0"'''
    assert str(Angle(hours=(12, 30))) == '''12h 30m 00.00s'''

    assert str(Angle(degrees=(90, 15, 30))) == '''90deg 15' 30.0"'''
    assert str(Angle(hours=(12, 30, 15))) == '''12h 30m 15.00s'''

def test_arcminutes_and_arcseconds_and_mas():
    angle = Angle(degrees=1.0)
    assert angle.arcminutes() == 60
    assert angle.arcseconds() == 60 * 60
    assert angle.mas() == 60 * 60 * 1000

def test_velocity_input_units():
    v1 = Velocity(au_per_d=2.0)
    v2 = Velocity(km_per_s=3462.9137)
    assert abs(v1.au_per_d - v2.au_per_d) < 1e-7

def test_stringifying_vector_distance():
    a = array([1.23, 4.56])
    s = str(Distance(au=a))
    if '[1' in s:
        # Python 3.5, says Travis CI.  No idea.
        assert s == '[1.23 4.56] au'
    else:
        # Every other version of Python.
        assert s == '[ 1.23  4.56] au'

def test_iterating_over_raw_measurement():
    distance = Distance(au=1.234)
    with assert_raises(UnpackingError) as a:
        x, y, z = distance
    assert str(a.exception) == '''\
cannot directly unpack a Distance into several values

To unpack a Distance into three components, you need to ask for its
value in specific units through an attribute or method:

    x, y, z = distance.au
    x, y, z = distance.km
    x, y, z = distance.to(astropy_unit)
'''

def test_iterating_over_raw_velocity():
    velocity = Velocity(au_per_d=1.234)
    with assert_raises(UnpackingError) as a:
        x, y, z = velocity
    assert str(a.exception) == '''\
cannot directly unpack a Velocity into several values

To unpack a Velocity into three components, you need to ask for its
value in specific units through an attribute or method:

    xdot, ydot, zdot = velocity.au_per_d
    xdot, ydot, zdot = velocity.km_per_s
    xdot, ydot, zdot = velocity.to(astropy_unit)
'''

def test_iterating_over_raw_angle():
    angle = Angle(degrees=4.5)
    with assert_raises(ValueError) as a:
        iter(angle)
    assert str(a.exception) == '''choose a specific Angle unit to iterate over

Instead of iterating over this Angle object, try iterating over one of
its unit-specific arrays like .degrees, .hours, or .radians, or else over
the output of one of its methods like .hstr(), .dstr(), .arcminutes(),
.arcseconds(), or .mas().  For all of the possibilities see:
https://rhodesmill.org/skyfield/api-units.html#skyfield.units.Angle'''

def test_converting_from_km_to_m():
    distance = Distance(km=1.234)
    assert abs(distance.m - 1234.0) < 1e-15

def test_converting_from_m_to_km():
    distance = Distance(m=1234.0)
    assert abs(distance.km - 1.234) < 1e-15

@needs_astropy
def test_converting_distance_with_astropy():
    distance = Distance(au=1.234)
    value1 = distance.km
    value2 = distance.to(u.km)
    epsilon = 0.02         # definitions of AU seem to disagree slightly
    assert abs(value1 - value2.value) < epsilon

@needs_astropy
def test_converting_velocity_with_astropy():
    velocity = Velocity(au_per_d=1.234)
    value1 = velocity.km_per_s
    value2 = velocity.to(u.km / u.s)
    epsilon = 1e-6
    assert abs(value1 - value2.value) < epsilon



import locale
import numpy as np
from datetime import datetime, timedelta
from threading import Lock

from .timelib import julian_date

_lock = Lock()

def _build_legacy_data(deltat_data, deltat_preds, leap_second_dat):
    data_end_time = deltat_data[0, -1]
    i = np.searchsorted(deltat_preds[0], data_end_time, side='right')
    delta_t_recent = np.concatenate([deltat_data, deltat_preds[:,i:]], axis=1)
    leap_dates, leap_offsets = leap_second_dat
    return delta_t_recent, leap_dates, leap_offsets

def parse_deltat_data(fileobj):
    """Parse the United States Naval Observatory ``deltat.data`` file.

    Each line file gives the date and the value of Delta T::

    2016  2  1  68.1577

    This function returns a 2xN array of raw Julian dates and matching
    Delta T values.

    """
    array = np.loadtxt(fileobj)
    year, month, day, delta_t = array.T
    return np.array((julian_date(year, month, day), delta_t))

def parse_deltat_preds(fileobj):
    """Parse the United States Naval Observatory ``deltat.preds`` file.

    The old format supplies a floating point year, the value of Delta T,
    and one or two other fields::

    2015.75      67.97               0.210         0.02

    The new format adds a modified Julian day as the first field:

    58484.000  2019.00   69.34      -0.152       0.117

    This function returns a 2xN array of raw Julian dates and matching
    Delta T values.

    """
    lines = iter(fileobj)
    header = next(lines)

    if header.startswith(b'YEAR'):
        # Format in use until 2019 February
        next(lines)             # discard blank line
        year_float, delta_t = np.loadtxt(lines, usecols=[0, 1]).T
    else:
        # Format in use since 2019 February
        year_float, delta_t = np.loadtxt(lines, usecols=[1, 2]).T

    year = year_float.astype(int)
    month = 1 + (year_float * 12.0).astype(int) % 12
    return np.array((julian_date(year, month, 1), delta_t))

def parse_leap_seconds(fileobj):
    """Parse the IERS file ``Leap_Second.dat``.

    The leap dates array can be searched with::

        index = np.searchsorted(leap_dates, jd, 'right')

    The resulting index allows (TAI - UTC) to be fetched with::

        offset = leap_offsets[index]

    """
    lines = iter(fileobj)
    for line in lines:
        if line.startswith(b'#  File expires on'):
            break
    else:
        raise ValueError('Leap_Second.dat is missing its expiration date')
    line = line.decode('ascii')

    with _lock:
        original_locale = locale.setlocale(locale.LC_ALL)
        locale.setlocale(locale.LC_ALL, 'C')
        try:
            dt = datetime.strptime(line, '#  File expires on %d %B %Y\n')
        finally:
            locale.setlocale(locale.LC_ALL, original_locale)

    # The file went out of date at the beginning of July 2016, and kept
    # downloading every time a user ran a Skyfield program.  So we now
    # build in a grace period:
    grace_period = timedelta(days=30)
    expiration_date = dt.date() + grace_period
    mjd, day, month, year, offsets = np.loadtxt(lines).T
    leap_dates = np.ndarray(len(mjd) + 2)
    leap_dates[0] = '-inf'
    leap_dates[1:-1] = mjd + 2400000.5
    leap_dates[-1] = 'inf'
    leap_offsets = np.ndarray(len(mjd) + 2)
    leap_offsets[0] = leap_offsets[1] = offsets[0]
    leap_offsets[2:] = offsets
    return expiration_date, (leap_dates, leap_offsets)
